open Util;;
open Time;;
open Event;;
open Vec;;


        (** constants **)

let pi = 4.0 *. atan(1.0)
let subunity = 1.0 -. (2.0 ** (-50.0))

        (** arithmetic **)

let dot x y = 
  V.foldl (+.) 0.0 (V.map2 ( *.) x y)
let sinc x = 
  if x=0.0 then 1.0 else let pix = pi*.x in (sin pix)/.pix

let amp2db a = 20.0 *. (log10 a)
let db2amp d = 10.0 ** (d /. 20.0)

        (** windows **)

let f_hamming x =
  0.54 +. (0.46 *. cos(pi*.x))
let f_blackman3 x =
  0.42 +. (0.5 *. cos(pi*.x)) +. (0.08 *. cos(2.0 *. pi*.x))
let f_blackman4 x =
  0.35875 +. (0.48829 *. cos(pi*.x)) +. (0.14128 *. cos(2.0 *. pi*.x)) 
    +. (0.01168 *. cos(3.0 *. pi*.x))
let window f n =
  let nf = float n
  in let x = V.count_by (-1.0) (2.0 /. (nf -. 1.0)) n
  in V.map f x

let hamming   = window f_hamming        (* -40 dB *)
let blackman3 = window f_blackman3      (* -58 dB *)
let blackman4 = window f_blackman4      (* -92 dB *)

        (** interpolation **)

(*
let linear_resample x freq phase0 =
  let t = V.count_by_upto phase0 freq (float ((V.length x) - 1))
  in let i = V.map int_of_float t
  in V.map2 
    (fun t i ->
      let a = t -. (float i) 
      in (1.0-.a) *. (V.peek i x) +. a *. (V.peek (i+1) x) )
    t
    i
*)

let interp1 t x =
  let a = frac t
  and i = int t
  in (1.0-.a) *. (V.peek i x) +. a *. (V.peek (i+1) x)

let linear_resample x freq phase0 =
  let t = V.count_by_upto phase0 freq (float ((V.length x) - 1))
  in V.map (fun t -> interp1 t x) t

        (** phasors **)

(* requires 0 <= f *)
let phasor n f ph0 = 
  let step ph = mod_float (ph +. f) 1.0
  in V.iterate n step ph0

(* requires 0 <= f < 1 *)
let phasor_wrap tmax f ph0 =
  let i = ref 0
  and ph = ref ph0
  and tv = ref [||]
  in (
    while (!i < tmax) do 
      let fi = (V.peek !i f) in 
      (
       if (!ph >= 1.0) then (
         fdec 1.0 ph;
         tv := V.join !tv (V.vec ((float !i) -. (!ph /. fi)))
        ) else ();
       incr i;
       finc fi ph;
      );
    done;
    !tv
 )

        (** tables **)

let tablei tab ph = 
  let ntab = subunity *. (float ((V.length tab) - 1))
  in V.map (fun ph -> interp1 (ntab *. ph) tab) ph

        (** oscs **)

(* 0 <= f < 0.5
   0 <= ph0 < 1 *)
let osc_sine n f ph0 = 
  let twopif   = 2.0*.pi*.f
  and twopiph0 = 2.0*.pi*.ph0
  in let twoc = 2.0*.cos(twopif)
     and x0 = sin twopiph0
     and x1 = sin (twopiph0 -. twopif)
     and x = V.const n 0.0
  in let rec loop i x0 x1 =
    (V.poke i x0 x;
     if (i < n-1) then (loop (i+1) (twoc*.x0 -. x1) x0) else ()
    )
  in (loop 0 x0 x1;
      x)

(* 0 <= f < 0.5 *)
let osc_sine' n f ph0 = 
  let t = V.count_by ph0 f n
  in V.map (fun t -> sin (2.0*.pi*.t)) t

(* 0 <= f < 0.5 *)
let osc_additive n freq_amp_list = 
  let bank = V.map (fun (f,a) -> V.scale a (osc_sine n f 0.0)) freq_amp_list
  in V.foldl (fun sum x -> V.map2 (+.) sum x) (V.const n 0.0) bank

(* 0 <= f < 0.5 *)
let osci tab n f ph0 =
  tablei tab (phasor n f ph0)

let noise_white n = 
  V.table n (fun _ -> (Random.float 2.0) -. 1.0)

        (** filters **)

let lowpass1 fc x =
  let b = 2.0 -. cos(pi*.fc)
  in let alpha = b -. sqrt(b**2.0 -. 1.0)
  in V.comp1i 
    (fun i ax -> ax*.alpha +. ((V.peek i x) *. (1.0 -. alpha)))
    (V.CompIter (0.0, V.length x))

let lowpass2 fc x =
  lowpass1 fc (lowpass1 fc x)

let lowpass4 fc x =
  lowpass2 fc (lowpass2 fc x)

        (** control **)

let vibrato n f amp pmix apmix = 
  let  periodic = osc_sine n f 0.0
  and aperiodic = V.scale (1.0/.sqrt(f)) (lowpass4 (2.0*.f) (noise_white n))
  in V.map2 (fun p ap -> amp *. (pmix*.p +. apmix*.ap))
    periodic aperiodic
